"""
    zeep.xsd
    --------

"""
from zeep.xsd.const import Nil, SkipValue  # noqa
from zeep.xsd.elements import *  # noqa
from zeep.xsd.schema import Schema  # noqa
from zeep.xsd.types import *  # noqa
from zeep.xsd.types.builtins import *  # noqa
from zeep.xsd.valueobjects import *  # noqa



import logging
import os
from contextlib import contextmanager
from urllib.parse import urlparse

import requests
from requests import Response
from requests_file import FileAdapter

from zeep.exceptions import TransportError
from zeep.utils import get_media_type, get_version
from zeep.wsdl.utils import etree_to_string

try:
    import httpx
except ImportError:
    httpx = None


__all__ = ["AsyncTransport", "Transport"]


class Transport:
    """The transport object handles all communication to the SOAP server.

    :param cache: The cache object to be used to cache GET requests
    :param timeout: The timeout for loading wsdl and xsd documents.
    :param operation_timeout: The timeout for operations (POST/GET). By
                              default this is None (no timeout).
    :param session: A :py:class:`request.Session()` object (optional)

    """

    def __init__(self, cache=None, timeout=300, operation_timeout=None, session=None):
        self.cache = cache
        self.load_timeout = timeout
        self.operation_timeout = operation_timeout
        self.logger = logging.getLogger(__name__)

        self.session = session or requests.Session()
        self.session.mount("file://", FileAdapter())
        self.session.headers["User-Agent"] = "Zeep/%s (www.python-zeep.org)" % (
            get_version()
        )

    def get(self, address, params, headers):
        """Proxy to requests.get()

        :param address: The URL for the request
        :param params: The query parameters
        :param headers: a dictionary with the HTTP headers.

        """
        response = self.session.get(
            address, params=params, headers=headers, timeout=self.operation_timeout
        )
        return response

    def post(self, address, message, headers):
        """Proxy to requests.posts()

        :param address: The URL for the request
        :param message: The content for the body
        :param headers: a dictionary with the HTTP headers.

        """
        if self.logger.isEnabledFor(logging.DEBUG):
            log_message = message
            if isinstance(log_message, bytes):
                log_message = log_message.decode("utf-8")
            self.logger.debug("HTTP Post to %s:\n%s", address, log_message)

        response = self.session.post(
            address, data=message, headers=headers, timeout=self.operation_timeout
        )

        if self.logger.isEnabledFor(logging.DEBUG):
            media_type = get_media_type(
                response.headers.get("Content-Type", "text/xml")
            )

            if media_type == "multipart/related":
                log_message = response.content
            else:
                log_message = response.content
                if isinstance(log_message, bytes):
                    log_message = log_message.decode(response.encoding or "utf-8")

            self.logger.debug(
                "HTTP Response from %s (status: %d):\n%s",
                address,
                response.status_code,
                log_message,
            )

        return response

    def post_xml(self, address, envelope, headers):
        """Post the envelope xml element to the given address with the headers.

        This method is intended to be overriden if you want to customize the
        serialization of the xml element. By default the body is formatted
        and encoded as utf-8. See ``zeep.wsdl.utils.etree_to_string``.

        """
        message = etree_to_string(envelope)
        return self.post(address, message, headers)

    def load(self, url):
        """Load the content from the given URL"""
        if not url:
            raise ValueError("No url given to load")

        scheme = urlparse(url).scheme
        if scheme in ("http", "https", "file"):

            if self.cache:
                response = self.cache.get(url)
                if response:
                    return bytes(response)

            content = self._load_remote_data(url)

            if self.cache:
                self.cache.add(url, content)

            return content
        else:
            with open(os.path.expanduser(url), "rb") as fh:
                return fh.read()

    def _load_remote_data(self, url):
        self.logger.debug("Loading remote data from: %s", url)
        response = self.session.get(url, timeout=self.load_timeout)
        response.raise_for_status()
        return response.content

    @contextmanager
    def settings(self, timeout=None):
        """Context manager to temporarily overrule options.

        Example::

            transport = zeep.Transport()
            with transport.settings(timeout=10):
                client.service.fast_call()

        :param timeout: Set the timeout for POST/GET operations (not used for
                        loading external WSDL or XSD documents)

        """
        old_timeout = self.operation_timeout
        self.operation_timeout = timeout
        yield
        self.operation_timeout = old_timeout


class AsyncTransport(Transport):
    """Asynchronous Transport class using httpx.

    Note that loading the wsdl is still a sync process since and only the
    operations can be called via async.

    """

    def __init__(
        self,
        client=None,
        wsdl_client=None,
        cache=None,
        timeout=300,
        operation_timeout=None,
        session=None,
        verify_ssl=True,
        proxy=None,
    ):
        if httpx is None:
            raise RuntimeError("The AsyncTransport is based on the httpx module")

        self.cache = cache
        self.wsdl_client = wsdl_client or httpx.Client(
            verify=verify_ssl,
            proxies=proxy,
            timeout=timeout,
        )
        self.client = client or httpx.AsyncClient(
            verify=verify_ssl,
            proxies=proxy,
            timeout=operation_timeout,
        )
        self.logger = logging.getLogger(__name__)

        self.wsdl_client.headers = {
            "User-Agent": "Zeep/%s (www.python-zeep.org)" % (get_version())
        }
        self.client.headers = {
            "User-Agent": "Zeep/%s (www.python-zeep.org)" % (get_version())
        }

    async def aclose(self):
        await self.client.aclose()

    def _load_remote_data(self, url):
        response = self.wsdl_client.get(url)
        result = response.read()

        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            raise TransportError(status_code=response.status_code)
        return result

    async def post(self, address, message, headers):
        self.logger.debug("HTTP Post to %s:\n%s", address, message)
        response = await self.client.post(
            address,
            data=message,
            headers=headers,
        )
        self.logger.debug(
            "HTTP Response from %s (status: %d):\n%s",
            address,
            response.status_code,
            response.read(),
        )
        return response

    async def post_xml(self, address, envelope, headers):
        message = etree_to_string(envelope)
        response = await self.post(address, message, headers)
        return self.new_response(response)

    async def get(self, address, params, headers):
        response = await self.client.get(
            address,
            params=params,
            headers=headers,
        )
        return self.new_response(response)

    def new_response(self, response):
        """Convert an aiohttp.Response object to a requests.Response object"""
        body = response.read()

        new = Response()
        new._content = body
        new.status_code = response.status_code
        new.headers = response.headers
        new.cookies = response.cookies
        new.encoding = response.encoding
        return new



import logging
import typing

from lxml import etree
from requests_toolbelt.multipart.decoder import MultipartDecoder

from zeep import ns, plugins, wsa
from zeep.exceptions import Fault, TransportError, XMLSyntaxError
from zeep.loader import parse_xml
from zeep.utils import as_qname, get_media_type, qname_attr
from zeep.wsdl.attachments import MessagePack
from zeep.wsdl.definitions import Binding, Operation
from zeep.wsdl.messages import DocumentMessage, RpcMessage
from zeep.wsdl.messages.xop import process_xop
from zeep.wsdl.utils import etree_to_string, url_http_to_https

if typing.TYPE_CHECKING:
    from zeep.wsdl.wsdl import Definition


logger = logging.getLogger(__name__)


class SoapBinding(Binding):
    """Soap 1.1/1.2 binding"""

    def __init__(self, wsdl, name, port_name, transport, default_style):
        """The SoapBinding is the base class for the Soap11Binding and
        Soap12Binding.

        :param wsdl:
        :type wsdl:
        :param name:
        :type name: string
        :param port_name:
        :type port_name: string
        :param transport:
        :type transport: zeep.transports.Transport
        :param default_style:

        """
        super().__init__(wsdl, name, port_name)
        self.transport = transport
        self.default_style = default_style

    @classmethod
    def match(cls, node):
        """Check if this binding instance should be used to parse the given
        node.

        :param node: The node to match against
        :type node: lxml.etree._Element

        """
        soap_node = node.find("soap:binding", namespaces=cls.nsmap)
        return soap_node is not None

    def create_message(self, operation, *args, **kwargs):
        envelope, http_headers = self._create(operation, args, kwargs)
        return envelope

    def _create(self, operation, args, kwargs, client=None, options=None):
        """Create the XML document to send to the server.

        Note that this generates the soap envelope without the wsse applied.

        """
        operation_obj = self.get(operation)
        if not operation_obj:
            raise ValueError("Operation %r not found" % operation)

        # Create the SOAP envelope
        serialized = operation_obj.create(*args, **kwargs)
        self._set_http_headers(serialized, operation_obj)

        envelope = serialized.content
        http_headers = serialized.headers

        # Apply ws-addressing
        if client:
            if not options:
                options = client.service._binding_options

            if operation_obj.abstract.wsa_action:
                envelope, http_headers = wsa.WsAddressingPlugin().egress(
                    envelope, http_headers, operation_obj, options
                )

            # Apply plugins
            envelope, http_headers = plugins.apply_egress(
                client, envelope, http_headers, operation_obj, options
            )

            # Apply WSSE
            if client.wsse:
                if isinstance(client.wsse, list):
                    for wsse in client.wsse:
                        envelope, http_headers = wsse.apply(envelope, http_headers)
                else:
                    envelope, http_headers = client.wsse.apply(envelope, http_headers)

        # Add extra http headers from the setings object
        if client.settings.extra_http_headers:
            http_headers.update(client.settings.extra_http_headers)

        return envelope, http_headers

    def send(self, client, options, operation, args, kwargs):
        """Called from the service

        :param client: The client with which the operation was called
        :type client: zeep.client.Client
        :param options: The binding options
        :type options: dict
        :param operation: The operation object from which this is a reply
        :type operation: zeep.wsdl.definitions.Operation
        :param args: The args to pass to the operation
        :type args: tuple
        :param kwargs: The kwargs to pass to the operation
        :type kwargs: dict

        """
        envelope, http_headers = self._create(
            operation, args, kwargs, client=client, options=options
        )

        response = client.transport.post_xml(options["address"], envelope, http_headers)

        operation_obj = self.get(operation)

        # If the client wants to return the raw data then let's do that.
        if client.settings.raw_response:
            return response

        return self.process_reply(client, operation_obj, response)

    async def send_async(self, client, options, operation, args, kwargs):
        """Called from the async service

        :param client: The client with which the operation was called
        :type client: zeep.client.Client
        :param options: The binding options
        :type options: dict
        :param operation: The operation object from which this is a reply
        :type operation: zeep.wsdl.definitions.Operation
        :param args: The args to pass to the operation
        :type args: tuple
        :param kwargs: The kwargs to pass to the operation
        :type kwargs: dict

        """
        envelope, http_headers = self._create(
            operation, args, kwargs, client=client, options=options
        )

        response = await client.transport.post_xml(
            options["address"], envelope, http_headers
        )

        if client.settings.raw_response:
            return response

        operation_obj = self.get(operation)
        return self.process_reply(client, operation_obj, response)

    def process_reply(self, client, operation, response):
        """Process the XML reply from the server.

        :param client: The client with which the operation was called
        :type client: zeep.client.Client
        :param operation: The operation object from which this is a reply
        :type operation: zeep.wsdl.definitions.Operation
        :param response: The response object returned by the remote server
        :type response: requests.Response

        """
        if response.status_code in (201, 202) and not response.content:
            return None

        elif response.status_code != 200 and not response.content:
            raise TransportError(
                u"Server returned HTTP status %d (no content available)"
                % response.status_code,
                status_code=response.status_code,
            )

        content_type = response.headers.get("Content-Type", "text/xml")
        media_type = get_media_type(content_type)
        message_pack = None

        # If the reply is a multipart/related then we need to retrieve all the
        # parts
        if media_type == "multipart/related":
            decoder = MultipartDecoder(
                response.content, content_type, response.encoding or "utf-8"
            )
            content = decoder.parts[0].content
            if len(decoder.parts) > 1:
                message_pack = MessagePack(parts=decoder.parts[1:])
        else:
            content = response.content

        try:
            doc = parse_xml(content, self.transport, settings=client.settings)
        except XMLSyntaxError as exc:
            raise TransportError(
                "Server returned response (%s) with invalid XML: %s.\nContent: %r"
                % (response.status_code, exc, response.content),
                status_code=response.status_code,
                content=response.content,
            )

        # Check if this is an XOP message which we need to decode first
        if message_pack:
            if process_xop(doc, message_pack):
                message_pack = None

        if client.wsse:
            client.wsse.verify(doc)

        doc, http_headers = plugins.apply_ingress(
            client, doc, response.headers, operation
        )

        # If the response code is not 200 or if there is a Fault node available
        # then assume that an error occured.
        fault_node = doc.find("soap-env:Body/soap-env:Fault", namespaces=self.nsmap)
        if response.status_code != 200 or fault_node is not None:
            return self.process_error(doc, operation)

        result = operation.process_reply(doc)

        if message_pack:
            message_pack._set_root(result)
            return message_pack
        return result

    def process_error(self, doc, operation):
        raise NotImplementedError

    def process_service_port(self, xmlelement, force_https=False):
        address_node = xmlelement.find("soap:address", namespaces=self.nsmap)
        if address_node is None:
            logger.debug("No valid soap:address found for service")
            return

        # Force the usage of HTTPS when the force_https boolean is true
        location = address_node.get("location")
        if force_https and location:
            location = url_http_to_https(location)
            if location != address_node.get("location"):
                logger.warning("Forcing soap:address location to HTTPS")

        return {"address": location}

    @classmethod
    def parse(cls, definitions, xmlelement):
        """

        Definition::

            <wsdl:binding name="nmtoken" type="qname"> *
                <-- extensibility element (1) --> *
                <wsdl:operation name="nmtoken"> *
                   <-- extensibility element (2) --> *
                   <wsdl:input name="nmtoken"? > ?
                       <-- extensibility element (3) -->
                   </wsdl:input>
                   <wsdl:output name="nmtoken"? > ?
                       <-- extensibility element (4) --> *
                   </wsdl:output>
                   <wsdl:fault name="nmtoken"> *
                       <-- extensibility element (5) --> *
                   </wsdl:fault>
                </wsdl:operation>
            </wsdl:binding>
        """
        name = qname_attr(xmlelement, "name", definitions.target_namespace)
        port_name = qname_attr(xmlelement, "type", definitions.target_namespace)

        # The soap:binding element contains the transport method and
        # default style attribute for the operations.
        soap_node = xmlelement.find("soap:binding", namespaces=cls.nsmap)
        transport = soap_node.get("transport")

        supported_transports = [
            "http://schemas.xmlsoap.org/soap/http",
            "http://www.w3.org/2003/05/soap/bindings/HTTP/",
        ]

        if transport not in supported_transports:
            raise NotImplementedError(
                "The binding transport %s is not supported (only soap/http)"
                % (transport)
            )
        default_style = soap_node.get("style", "document")

        obj = cls(definitions.wsdl, name, port_name, transport, default_style)
        for node in xmlelement.findall("wsdl:operation", namespaces=cls.nsmap):
            operation = SoapOperation.parse(definitions, node, obj, nsmap=cls.nsmap)
            obj._operation_add(operation)
        return obj


class Soap11Binding(SoapBinding):
    nsmap = {
        "soap": ns.SOAP_11,
        "soap-env": ns.SOAP_ENV_11,
        "wsdl": ns.WSDL,
        "xsd": ns.XSD,
    }

    def process_error(self, doc, operation):
        fault_node = doc.find("soap-env:Body/soap-env:Fault", namespaces=self.nsmap)

        if fault_node is None:
            raise Fault(
                message="Unknown fault occured",
                code=None,
                actor=None,
                detail=etree_to_string(doc),
            )

        def get_text(name):
            child = fault_node.find(name, namespaces=fault_node.nsmap)
            if child is not None:
                return child.text

        raise Fault(
            message=get_text("faultstring"),
            code=get_text("faultcode"),
            actor=get_text("faultactor"),
            detail=fault_node.find("detail", namespaces=fault_node.nsmap),
        )

    def _set_http_headers(self, serialized, operation):
        serialized.headers["Content-Type"] = "text/xml; charset=utf-8"


class Soap12Binding(SoapBinding):
    nsmap = {
        "soap": ns.SOAP_12,
        "soap-env": ns.SOAP_ENV_12,
        "wsdl": ns.WSDL,
        "xsd": ns.XSD,
    }

    def process_error(self, doc, operation):
        fault_node = doc.find("soap-env:Body/soap-env:Fault", namespaces=self.nsmap)

        if fault_node is None:
            raise Fault(
                message="Unknown fault occured",
                code=None,
                actor=None,
                detail=etree_to_string(doc),
            )

        def get_text(name):
            child = fault_node.find(name)
            if child is not None:
                return child.text

        message = fault_node.findtext(
            "soap-env:Reason/soap-env:Text", namespaces=self.nsmap
        )
        code = fault_node.findtext(
            "soap-env:Code/soap-env:Value", namespaces=self.nsmap
        )

        # Extract the fault subcodes. These can be nested, as in subcodes can
        # also contain other subcodes.
        subcodes = []
        subcode_element = fault_node.find(
            "soap-env:Code/soap-env:Subcode", namespaces=self.nsmap
        )
        while subcode_element is not None:
            subcode_value_element = subcode_element.find(
                "soap-env:Value", namespaces=self.nsmap
            )
            subcode_qname = as_qname(
                subcode_value_element.text, subcode_value_element.nsmap, None
            )
            subcodes.append(subcode_qname)
            subcode_element = subcode_element.find(
                "soap-env:Subcode", namespaces=self.nsmap
            )

        # TODO: We should use the fault message as defined in the wsdl.
        detail_node = fault_node.find("soap-env:Detail", namespaces=self.nsmap)
        raise Fault(
            message=message,
            code=code,
            actor=None,
            detail=detail_node,
            subcodes=subcodes,
        )

    def _set_http_headers(self, serialized, operation):
        serialized.headers["Content-Type"] = "; ".join(
            [
                "application/soap+xml",
                "charset=utf-8",
                'action="%s"' % operation.soapaction,
            ]
        )


class SoapOperation(Operation):
    """Represent's an operation within a specific binding."""

    def __init__(self, name, binding, nsmap, soapaction, style):
        super().__init__(name, binding)
        self.nsmap = nsmap
        self.soapaction = soapaction
        self.style = style

    def process_reply(self, envelope):
        envelope_qname = etree.QName(self.nsmap["soap-env"], "Envelope")
        if envelope.tag != envelope_qname:
            raise XMLSyntaxError(
                (
                    "The XML returned by the server does not contain a valid "
                    + "{%s}Envelope root element. The root element found is %s "
                )
                % (envelope_qname.namespace, envelope.tag)
            )

        if self.output:
            return self.output.deserialize(envelope)

    @classmethod
    def parse(cls, definitions, xmlelement, binding, nsmap):
        """

        Definition::

            <wsdl:operation name="nmtoken"> *
                <soap:operation soapAction="uri"? style="rpc|document"?>?
                <wsdl:input name="nmtoken"? > ?
                    <soap:body use="literal"/>
               </wsdl:input>
               <wsdl:output name="nmtoken"? > ?
                    <-- extensibility element (4) --> *
               </wsdl:output>
               <wsdl:fault name="nmtoken"> *
                    <-- extensibility element (5) --> *
               </wsdl:fault>
            </wsdl:operation>

        Example::

            <wsdl:operation name="GetLastTradePrice">
              <soap:operation soapAction="http://example.com/GetLastTradePrice"/>
              <wsdl:input>
                <soap:body use="literal"/>
              </wsdl:input>
              <wsdl:output>
              </wsdl:output>
              <wsdl:fault name="dataFault">
                <soap:fault name="dataFault" use="literal"/>
              </wsdl:fault>
            </operation>

        """
        name = xmlelement.get("name")

        # The soap:operation element is required for soap/http bindings
        # and may be omitted for other bindings.
        soap_node = xmlelement.find("soap:operation", namespaces=binding.nsmap)
        action = None
        if soap_node is not None:
            action = soap_node.get("soapAction")
            style = soap_node.get("style", binding.default_style)
        else:
            style = binding.default_style

        obj = cls(name, binding, nsmap, action, style)

        if style == "rpc":
            message_class = RpcMessage
        else:
            message_class = DocumentMessage

        for node in xmlelement:
            tag_name = etree.QName(node.tag).localname
            if tag_name not in ("input", "output", "fault"):
                continue
            msg = message_class.parse(
                definitions=definitions,
                xmlelement=node,
                operation=obj,
                nsmap=nsmap,
                type=tag_name,
            )
            if tag_name == "fault":
                obj.faults[msg.name] = msg
            else:
                setattr(obj, tag_name, msg)

        return obj

    def resolve(self, definitions: "Definition"):
        super().resolve(definitions)
        for name, fault in self.faults.items():
            if name in self.abstract.fault_messages:
                fault.resolve(definitions, self.abstract.fault_messages[name])

        if self.output:
            self.output.resolve(definitions, self.abstract.output_message)
        if self.input:
            self.input.resolve(definitions, self.abstract.input_message)



import typing

from lxml import etree

from zeep.xsd.types.base import Type
from zeep.xsd.types.collection import UnionType  # FIXME
from zeep.xsd.types.simple import AnySimpleType  # FIXME

if typing.TYPE_CHECKING:
    from zeep.xsd.types.complex import ComplexType
    from zeep.xsd.valueobjects import CompoundValue


class UnresolvedType(Type):
    def __init__(self, qname, schema):
        self.qname = qname
        assert self.qname.text != "None"
        self.schema = schema

    def __repr__(self):
        return "<%s(qname=%r)>" % (self.__class__.__name__, self.qname.text)

    def render(
        self,
        node: etree._Element,
        value: typing.Union[list, dict, "CompoundValue"],
        xsd_type: "ComplexType" = None,
        render_path=None,
    ) -> None:
        raise RuntimeError(
            "Unable to render unresolved type %s. This is probably a bug."
            % (self.qname)
        )

    def resolve(self):
        retval = self.schema.get_type(self.qname)
        return retval.resolve()


class UnresolvedCustomType(Type):
    def __init__(self, qname, base_type, schema):
        assert qname is not None
        self.qname = qname
        self.name = str(qname.localname)
        self.schema = schema
        self.base_type = base_type

    def __repr__(self):
        return "<%s(qname=%r, base_type=%r)>" % (
            self.__class__.__name__,
            self.qname.text,
            self.base_type,
        )

    def resolve(self):
        base = self.base_type
        base = base.resolve()

        cls_attributes = {"__module__": "zeep.xsd.dynamic_types"}

        if issubclass(base.__class__, UnionType):
            xsd_type = type(self.name, (base.__class__,), cls_attributes)
            return xsd_type(base.item_types)

        elif issubclass(base.__class__, AnySimpleType):
            xsd_type = type(self.name, (base.__class__,), cls_attributes)
            return xsd_type(self.qname)

        else:
            xsd_type = type(self.name, (base.base_class,), cls_attributes)
            return xsd_type(self.qname)



import logging
import typing
from collections import OrderedDict

from lxml import etree

from zeep import exceptions, ns
from zeep.loader import load_external
from zeep.settings import Settings
from zeep.xsd import const
from zeep.xsd import elements as xsd_elements
from zeep.xsd import types as xsd_types
from zeep.xsd.elements import builtins as xsd_builtins_elements
from zeep.xsd.types import builtins as xsd_builtins_types
from zeep.xsd.visitor import SchemaVisitor

logger = logging.getLogger(__name__)


class Schema:
    """A schema is a collection of schema documents."""

    def __init__(self, node=None, transport=None, location=None, settings=None):
        """
        :param node:
        :param transport:
        :param location:
        :param settings: The settings object

        """
        self.settings = settings or Settings()

        self._transport = transport

        self.documents = _SchemaContainer()
        self._prefix_map_auto = {}
        self._prefix_map_custom = {}

        self._load_default_documents()

        if not isinstance(node, list):
            nodes = [node] if node is not None else []
        else:
            nodes = node
        self.add_documents(nodes, location)

    def __repr__(self):
        main_doc = self.root_document
        if main_doc:
            return "<Schema(location=%r, tns=%r)>" % (
                main_doc._location,
                main_doc._target_namespace,
            )
        return "<Schema()>"

    @property
    def prefix_map(self):
        retval = {}
        retval.update(self._prefix_map_custom)
        retval.update(
            {k: v for k, v in self._prefix_map_auto.items() if v not in retval.values()}
        )
        return retval

    @property
    def root_document(self):
        return next((doc for doc in self.documents if not doc._is_internal), None)

    @property
    def is_empty(self):
        """Boolean to indicate if this schema contains any types or elements"""
        return all(document.is_empty for document in self.documents)

    @property
    def namespaces(self):
        return self.documents.get_all_namespaces()

    @property
    def elements(self):
        """Yield all globla xsd.Type objects

        :rtype: Iterable of zeep.xsd.Element

        """
        seen = set()
        for document in self.documents:
            for element in document._elements.values():
                if element.qname not in seen:
                    yield element
                    seen.add(element.qname)

    @property
    def types(self):
        """Yield all global xsd.Type objects

        :rtype: Iterable of zeep.xsd.ComplexType

        """
        seen = set()
        for document in self.documents:
            for type_ in document._types.values():
                if type_.qname not in seen:
                    yield type_
                    seen.add(type_.qname)

    def add_documents(
        self, schema_nodes: typing.List[etree._Element], location: str
    ) -> None:
        resolve_queue = []
        for node in schema_nodes:
            document = self.create_new_document(node, location)
            resolve_queue.append(document)

        for document in resolve_queue:
            document.resolve()

        self._prefix_map_auto = self._create_prefix_map()

    def add_document_by_url(self, url: str) -> None:
        schema_node = load_external(url, self._transport, settings=self.settings)
        document = self.create_new_document(schema_node, url=url)
        document.resolve()

    def get_element(self, qname) -> xsd_elements.Element:
        """Return a global xsd.Element object with the given qname"""
        qname = self._create_qname(qname)
        return self._get_instance(qname, "get_element", "element")

    def get_type(self, qname, fail_silently=False):
        """Return a global xsd.Type object with the given qname

        :rtype: zeep.xsd.ComplexType or zeep.xsd.AnySimpleType


        """
        qname = self._create_qname(qname)
        try:
            return self._get_instance(qname, "get_type", "type")
        except exceptions.NamespaceError as exc:
            if fail_silently:
                logger.debug(str(exc))
            else:
                raise

    def get_group(self, qname) -> xsd_elements.Group:
        """Return a global xsd.Group object with the given qname."""
        return self._get_instance(qname, "get_group", "group")

    def get_attribute(self, qname) -> xsd_elements.Attribute:
        """Return a global xsd.attribute object with the given qname"""
        return self._get_instance(qname, "get_attribute", "attribute")

    def get_attribute_group(self, qname) -> xsd_elements.AttributeGroup:
        """Return a global xsd.attributeGroup object with the given qname"""
        return self._get_instance(qname, "get_attribute_group", "attributeGroup")

    def set_ns_prefix(self, prefix, namespace):
        self._prefix_map_custom[prefix] = namespace

    def get_ns_prefix(self, prefix):
        try:
            try:
                return self._prefix_map_custom[prefix]
            except KeyError:
                return self._prefix_map_auto[prefix]
        except KeyError:
            raise ValueError("No such prefix %r" % prefix)

    def get_shorthand_for_ns(self, namespace):
        for prefix, other_namespace in self._prefix_map_auto.items():
            if namespace == other_namespace:
                return prefix
        for prefix, other_namespace in self._prefix_map_custom.items():
            if namespace == other_namespace:
                return prefix

        if namespace == "http://schemas.xmlsoap.org/soap/envelope/":
            return "soap-env"
        return namespace

    def create_new_document(self, node, url, base_url=None, target_namespace=None):
        """

        :rtype: zeep.xsd.schema.SchemaDocument

        """
        namespace = node.get("targetNamespace") if node is not None else None
        if not namespace:
            namespace = target_namespace
        if base_url is None:
            base_url = url

        schema = SchemaDocument(namespace, url, base_url)
        self.documents.add(schema)
        schema.load(self, node)
        return schema

    def merge(self, schema):
        """Merge an other XSD schema in this one"""
        for document in schema.documents:
            self.documents.add(document)
        self._prefix_map_auto = self._create_prefix_map()

    def deserialize(self, node):
        elm = self.get_element(node.tag)
        return elm.parse(node, schema=self)

    def _load_default_documents(self):
        schema = SchemaDocument(ns.XSD, None, None)

        for cls in xsd_builtins_types._types:
            instance = cls(is_global=True)
            schema.register_type(cls._default_qname, instance)

        for cls in xsd_builtins_elements._elements:
            instance = cls()
            schema.register_element(cls.qname, instance)

        schema._is_internal = True
        self.documents.add(schema)
        return schema

    def _get_instance(self, qname, method_name, name):
        """Return an object from one of the SchemaDocument's"""
        qname = self._create_qname(qname)
        try:
            last_exception = None  # type: typing.Optional[BaseException]
            for schema in self._get_schema_documents(qname.namespace):
                method = getattr(schema, method_name)
                try:
                    return method(qname)
                except exceptions.LookupError as exc:
                    last_exception = exc
                    continue
            if last_exception is not None:
                raise last_exception

        except exceptions.NamespaceError:
            raise exceptions.NamespaceError(
                (
                    "Unable to resolve %s %s. "
                    + "No schema available for the namespace %r."
                )
                % (name, qname.text, qname.namespace)
            )

    def _create_qname(self, name):
        """Create an `lxml.etree.QName()` object for the given qname string.

        This also expands the shorthand notation.

        :rtype: lxml.etree.QNaame

        """
        if isinstance(name, etree.QName):
            return name

        if not name.startswith("{") and ":" in name and self._prefix_map_auto:
            prefix, localname = name.split(":", 1)
            if prefix in self._prefix_map_custom:
                return etree.QName(self._prefix_map_custom[prefix], localname)
            elif prefix in self._prefix_map_auto:
                return etree.QName(self._prefix_map_auto[prefix], localname)
            else:
                raise ValueError("No namespace defined for the prefix %r" % prefix)
        else:
            return etree.QName(name)

    def _create_prefix_map(self):
        prefix_map = {"xsd": "http://www.w3.org/2001/XMLSchema"}
        i = 0
        for namespace in self.documents.get_all_namespaces():
            if namespace is None or namespace in prefix_map.values():
                continue

            prefix_map["ns%d" % i] = namespace
            i += 1
        return prefix_map

    def _get_schema_documents(self, namespace, fail_silently=False):
        """Return a list of SchemaDocument's for the given namespace.

        :rtype: list of SchemaDocument

        """
        if (
            not self.documents.has_schema_document_for_ns(namespace)
            and namespace in const.AUTO_IMPORT_NAMESPACES
        ):
            logger.debug("Auto importing missing known schema: %s", namespace)
            self.add_document_by_url(namespace)

        return self.documents.get_by_namespace(namespace, fail_silently)


class _SchemaContainer:
    """Container instances to store multiple SchemaDocument objects per
    namespace.

    """

    def __init__(self):
        self._instances = OrderedDict()

    def __iter__(self):
        for document in self.values():
            yield document

    def add(self, document: "SchemaDocument") -> None:
        """Add a schema document"""
        logger.debug(
            "Add document with tns %s to schema %s", document.namespace, id(self)
        )
        documents = self._instances.setdefault(document.namespace, [])
        documents.append(document)

    def get_all_namespaces(self) -> typing.List[str]:
        return list(self._instances.keys())

    def get_by_namespace(
        self, namespace, fail_silently
    ) -> typing.List["SchemaDocument"]:
        if namespace not in self._instances:
            if fail_silently:
                return []
            raise exceptions.NamespaceError(
                "No schema available for the namespace %r" % namespace
            )
        return self._instances[namespace]

    def get_by_namespace_and_location(
        self, namespace: str, location: str
    ) -> typing.Optional["SchemaDocument"]:
        """Return a SchemaDocument for the given namespace AND location"""
        documents = self.get_by_namespace(namespace, fail_silently=True)
        for document in documents:
            if document._location == location:
                return document
        return None

    def has_schema_document_for_ns(self, namespace: str) -> bool:
        """Return a boolean if there is a SchemaDocument for the namespace.

        :rtype: boolean

        """
        return namespace in self._instances

    def values(self):
        for documents in self._instances.values():
            for document in documents:
                yield document


class SchemaDocument:
    """A Schema Document consists of a set of schema components for a
    specific target namespace.

    This represents an xsd:Schema object

    """

    def __init__(self, namespace, location, base_url):
        logger.debug("Init schema document for %r", location)

        # Internal
        self._base_url = base_url or location
        self._location = location
        self._target_namespace = namespace
        self._is_internal = False
        self._has_empty_import = False

        # Containers for specific types
        self._attribute_groups = {}
        self._attributes = {}
        self._elements = {}
        self._groups = {}
        self._types = {}

        self._imports = OrderedDict()
        self._element_form = "unqualified"
        self._attribute_form = "unqualified"
        self._resolved = False
        # self._xml_schema = None

    def __repr__(self):
        return "<SchemaDocument(location=%r, tns=%r, is_empty=%r)>" % (
            self._location,
            self._target_namespace,
            self.is_empty,
        )

    @property
    def namespace(self):
        return self._target_namespace

    @property
    def is_empty(self):
        return not bool(self._imports or self._types or self._elements)

    def load(self, schema, node):
        """Load the XML Schema passed in via the node attribute.

        :type schema: zeep.xsd.schema.Schema
        :type node: etree._Element

        """
        if node is None:
            return

        if not schema.documents.has_schema_document_for_ns(self._target_namespace):
            raise RuntimeError(
                "The document needs to be registered in the schema before "
                + "it can be loaded"
            )

        # Disable XML schema validation for now
        # if len(node) > 0:
        #     self.xml_schema = etree.XMLSchema(node)
        visitor = SchemaVisitor(schema, self)
        visitor.visit_schema(node)

    def resolve(self):
        logger.debug("Resolving in schema %s", self)

        if self._resolved:
            return
        self._resolved = True

        for schemas in self._imports.values():
            for schema in schemas:
                schema.resolve()

        def _resolve_dict(val):
            try:
                for key, obj in val.items():
                    new = obj.resolve()
                    assert new is not None, "resolve() should return an object"
                    val[key] = new
            except exceptions.LookupError as exc:
                raise exceptions.LookupError(
                    (
                        "Unable to resolve %(item_name)s %(qname)s in "
                        "%(file)s. (via %(parent)s)"
                    )
                    % {
                        "item_name": exc.item_name,
                        "qname": exc.qname,
                        "file": exc.location,
                        "parent": obj.qname,
                    }
                )

        _resolve_dict(self._attribute_groups)
        _resolve_dict(self._attributes)
        _resolve_dict(self._elements)
        _resolve_dict(self._groups)
        _resolve_dict(self._types)

    def register_import(self, namespace, schema):
        """Register an import for an other schema document.

        :type namespace: str
        :type schema: zeep.xsd.schema.SchemaDocument

        """
        schemas = self._imports.setdefault(namespace, [])
        schemas.append(schema)

    def is_imported(self, namespace):
        return namespace in self._imports

    def register_type(self, qname: etree.QName, value: xsd_types.Type):
        """Register a xsd.Type in this schema"""
        self._add_component(qname, value, self._types, "type")

    def register_element(self, qname: etree.QName, value: xsd_elements.Element):
        """Register a xsd.Element in this schema"""
        self._add_component(qname, value, self._elements, "element")

    def register_group(self, qname: etree.QName, value: xsd_elements.Group):
        """Register a xsd:Group in this schema"""
        self._add_component(qname, value, self._groups, "group")

    def register_attribute(self, qname: str, value: xsd_elements.Attribute):
        """Register a xsd:Attribute in this schema"""
        self._add_component(qname, value, self._attributes, "attribute")

    def register_attribute_group(
        self, qname: etree.QName, value: xsd_elements.AttributeGroup
    ) -> None:
        """Register a xsd:AttributeGroup in this schema"""
        self._add_component(qname, value, self._attribute_groups, "attribute_group")

    def get_type(self, qname: etree.QName):
        """Return a xsd.Type object from this schema

        :rtype: zeep.xsd.ComplexType or zeep.xsd.AnySimpleType

        """
        return self._get_component(qname, self._types, "type")

    def get_element(self, qname) -> xsd_elements.Element:
        """Return a xsd.Element object from this schema"""
        return self._get_component(qname, self._elements, "element")

    def get_group(self, qname) -> xsd_elements.Group:
        """Return a xsd.Group object from this schema"""
        return self._get_component(qname, self._groups, "group")

    def get_attribute(self, qname) -> xsd_elements.Attribute:
        """Return a xsd.Attribute object from this schema"""
        return self._get_component(qname, self._attributes, "attribute")

    def get_attribute_group(self, qname) -> xsd_elements.AttributeGroup:
        """Return a xsd.AttributeGroup object from this schema"""
        return self._get_component(qname, self._attribute_groups, "attributeGroup")

    def _add_component(self, name, value, items, item_name):
        if isinstance(name, etree.QName):
            name = name.text
        logger.debug("register_%s(%r, %r)", item_name, name, value)
        items[name] = value

    def _get_component(self, qname, items, item_name):
        try:
            return items[qname]
        except KeyError:
            known_items = ", ".join(items.keys())
            raise exceptions.LookupError(
                (
                    "No %(item_name)s '%(localname)s' in namespace %(namespace)s. "
                    + "Available %(item_name_plural)s are: %(known_items)s"
                )
                % {
                    "item_name": item_name,
                    "item_name_plural": item_name + "s",
                    "localname": qname.localname,
                    "namespace": qname.namespace,
                    "known_items": known_items or " - ",
                },
                qname=qname,
                item_name=item_name,
                location=self._location,
            )



import cgi
import inspect
import typing

from lxml import etree

from zeep.exceptions import XMLParseError
from zeep.ns import XSD


def qname_attr(
    node: etree._Element,
    attr_name: typing.Union[str, etree.QName],
    target_namespace=None,
) -> typing.Optional[etree.QName]:
    value = node.get(attr_name)
    if value is not None:
        return as_qname(value, node.nsmap, target_namespace)
    return None


def as_qname(value: str, nsmap, target_namespace=None) -> etree.QName:
    """Convert the given value to a QName"""
    value = value.strip()  # some xsd's contain leading/trailing spaces
    if ":" in value:
        prefix, local = value.split(":")

        # The xml: prefix is always bound to the XML namespace, see
        # https://www.w3.org/TR/xml-names/
        if prefix == "xml":
            namespace = "http://www.w3.org/XML/1998/namespace"
        else:
            namespace = nsmap.get(prefix)

        if not namespace:
            raise XMLParseError("No namespace defined for %r (%r)" % (prefix, value))

        # Workaround for https://github.com/mvantellingen/python-zeep/issues/349
        if not local:
            return etree.QName(XSD, "anyType")

        return etree.QName(namespace, local)

    if target_namespace:
        return etree.QName(target_namespace, value)

    if nsmap.get(None):
        return etree.QName(nsmap[None], value)
    return etree.QName(value)


def findall_multiple_ns(node: etree._Element, name, namespace_sets):
    result = []
    for nsmap in namespace_sets:
        result.extend(node.findall(name, namespaces=nsmap))
    return result


def get_version():
    from zeep import __version__  # cyclic import

    return __version__


def get_base_class(objects):
    """Return the best base class for multiple objects.

    Implementation is quick and dirty, might be done better.. ;-)

    """
    bases = [inspect.getmro(obj.__class__)[::-1] for obj in objects]
    num_objects = len(objects)
    max_mro = max(len(mro) for mro in bases)

    base_class = None
    for i in range(max_mro):
        try:
            if len({bases[j][i] for j in range(num_objects)}) > 1:
                break
        except IndexError:
            break
        base_class = bases[0][i]
    return base_class


def detect_soap_env(envelope):
    root_tag = etree.QName(envelope)
    return root_tag.namespace


def get_media_type(value):
    """Parse a HTTP content-type header and return the media-type"""
    main_value, parameters = cgi.parse_header(value)
    return main_value.lower()
